//---------------------------------------------------------------------
//  Copyright (C) Microsoft Corporation.  All rights reserved.
// 
// THIS CODE AND INFORMATION ARE PROVIDED AS IS WITHOUT WARRANTY OF ANY
// KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
// IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A
// PARTICULAR PURPOSE.
//
//---------------------------------------------------------------------

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.ComponentModel.Design;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using System.Windows.Forms.Design.Behavior;

namespace DesignSurfaceExtenderLibrary.Extensions
{
    /// <summary>
    /// Provides an abstract base implementation for <see cref="Behavior"/> types
    /// which have designable rectangular bounds.
    /// </summary>
    /// <remarks>The <see cref="BoundsBehavior"/> class manages the details
    /// of drag/drop transactions. 
    /// </remarks>
    internal abstract class BoundsBehavior : Behavior
    {   
        /// <summary>
        /// Caches a reference to the control to which this
        /// <see cref="Behavior"/> is bound.
        /// </summary>
        protected Control relatedControl;

        /// <summary>
        /// The <see cref="IServiceProvider"/> which provides access to 
        /// designer services.
        /// </summary>
        private IServiceProvider serviceProvider;

        /// <summary>
        /// Backs the <see cref="BehaviorService"/> property.
        /// </summary>
        protected BehaviorService behaviorServiceValue;

        /// <summary>
        /// The <see cref="IDesignerHost"/> for creating transactions.
        /// </summary>
        private IDesignerHost designerHostValue;

        /// <summary>
        /// Tracks when a mouse capture behavior has been
        /// pushed onto the behavior stack.
        /// </summary>
        private bool pushedBehavior;

        /// <summary>
        /// Tracks when a mouse drag operation is active.
        /// </summary>
        private bool isDragging;

        /// <summary>
        /// The currently selected edge of the bounds glyph.
        /// </summary>
        protected EdgeSelection selectedEdge = EdgeSelection.None;

        /// <summary>
        /// Caches the position of the current drag operation.
        /// </summary>
        protected Point currentPosition;

        /// <summary>
        /// The transaction representing the current drag operation.
        /// </summary>
        private DesignerTransaction dragTransaction;

        /// <summary>
        /// The name of the drag transaction.
        /// </summary>
        protected string transactionName;

        /// <summary>
        /// The bounds-related property attached to this behavior.
        /// </summary>
        /// <remarks>The type-specific property name must be assigned by 
        /// deriving classes.
        /// </remarks>
        protected string propertyName;
        
        /// <summary>
        /// Initializes a new instance of the <see cref="BoundsBehavior"/> type.
        /// </summary>
        /// <param name="serviceProvider">An <see cref="IServiceProvider"/> which
        /// provides design-time services. </param>
        internal BoundsBehavior( IServiceProvider serviceProvider )
        {
            if( serviceProvider != null )
            {
                this.serviceProvider = serviceProvider;
            }
            else
            {
                throw new ArgumentNullException( "serviceProvider", "must be assigned" );
            }
        }

        /// <summary>
        /// Sets the type-specific bounds-related property by using the 
        /// given <see cref="PropertyDescriptor"/>.
        /// </summary>
        /// <param name="pd">The <see cref="PropertyDescriptor"/> to use for setting
        /// the bounds-related property.</param>
        /// <remarks>Deriving classes must provide the implementation for setting
        /// the bounds-related property specific to the extension type.
        /// </remarks>
        protected abstract void SetBoundsProperty( PropertyDescriptor pd );

        /// <summary>
        /// Gets a reference to the <see cref="BehaviorService"/>.
        /// </summary>
        internal BehaviorService BehaviorService
        {
            get
            {
                if( this.behaviorServiceValue == null )
                {   
                    this.behaviorServiceValue = this.serviceProvider.GetService( 
                        typeof( BehaviorService ) ) as BehaviorService;
                }

                return this.behaviorServiceValue;
            }
        }

        internal IDesignerHost DesignerHost
        {
            get
            {
                if( this.designerHostValue == null )
                {
                    this.designerHostValue = this.serviceProvider.GetService(
                        typeof( IDesignerHost ) ) as IDesignerHost;
                }

                return this.designerHostValue;
            }
        }

        public override bool OnMouseDown( 
            Glyph g, 
            MouseButtons button, 
            Point mouseLoc )
        {   
            if( button != MouseButtons.Left )
            {
                return this.pushedBehavior;
            }

            BoundsGlyph bg = g as BoundsGlyph;

            this.InitializeDragOperation( bg, mouseLoc );

            return false;
        }

        /// <summary>
        /// Initiates a drag operation. 
        /// </summary>
        /// <param name="bg">A <see cref="BoundsGlyph"/> which provides state
        /// for the drag operation.</param>
        /// <param name="mouseLoc">The location at which the drag operation starts.</param>
        /// <remarks>The <see cref="InitializeDragOperation"/> method pushes a mouse
        /// capture behavior onto the behavior stack. It also opens a designer transaction
        /// for the drag operation.
        /// </remarks>
        protected virtual void InitializeDragOperation( 
            BoundsGlyph bg,
            Point mouseLoc )
        {
            if( bg != null )
            {
                this.relatedControl = bg.RelatedControl;

                pushedBehavior = true;
                this.BehaviorService.PushCaptureBehavior( this );

                this.selectedEdge = bg.SelectedEdge( mouseLoc );

                this.OpenDragTransaction();
            }
        }

        public override bool OnMouseUp( Glyph g, MouseButtons button )
        {
            if( this.isDragging )
            {
                this.isDragging = false;

                this.CloseDragTransaction( true );

                this.OnLoseCapture( g, EventArgs.Empty );
            }

            return false;
        }

        public override void OnLoseCapture( Glyph g, EventArgs e )
        {   
            if( this.pushedBehavior )
            {
                this.pushedBehavior = false;

                if( this.BehaviorService != null )
                {
                    if( this.isDragging )
                    {
                        this.isDragging = false;

                        this.CloseDragTransaction( false );
                    }

                    this.selectedEdge = EdgeSelection.None;

                    this.BehaviorService.PopBehavior( this );
                }
            }
        }

        /// <summary>
        /// Creates a new transaction for a drag operation.
        /// </summary>
        protected virtual void OpenDragTransaction()
        {
            this.dragTransaction = 
                this.DesignerHost.CreateTransaction( this.transactionName );
        }

        /// <summary>
        /// Closes a drag transaction, with optional commit or cancel.
        /// </summary>
        /// <param name="commit"><c>true</c> to commit the transaction; 
        /// otherwise, <c>false</c>.</param>
        protected virtual void CloseDragTransaction( bool commit )
        {
            if( this.dragTransaction != null )
            {
                DesignerTransaction t = this.dragTransaction;
                this.dragTransaction = null;

                using( t )
                {
                    if( commit )
                    {
                        t.Commit();
                    }
                    else
                    {
                        t.Cancel();
                    }
                }
            }
        }

        public override bool OnMouseMove( Glyph g, MouseButtons button, Point mouseLoc )
        {   
            if( !this.pushedBehavior )
            {
                return false;
            }

            if( !this.isDragging )
            {
                this.isDragging = true;
            }

            this.currentPosition = mouseLoc;

            PropertyDescriptor pd = TypeDescriptor.GetProperties( this.relatedControl )[this.propertyName];

            this.SetBoundsProperty( pd );

            return true;
        }
    }
}
